/*
* Copyright 2004 - 2009 by dronten@gmail.com
*
* This source is distributed under the terms of the GNU PUBLIC LICENSE version 3
* http://www.gnu.org/licenses/gpl.html
*/

package com.googlepages.dronten.jtagger.audio;

import com.googlepages.dronten.jtagger.Constants;
import com.googlepages.dronten.jtagger.model.PictureModel;
import com.googlepages.dronten.jtagger.util.Log;
import com.googlepages.dronten.jtagger.util.Pref;
import com.googlepages.dronten.jtagger.util.Util;
import javazoom.jl.decoder.Bitstream;
import javazoom.jl.decoder.Header;
import org.blinkenlights.jid3.ID3Exception;
import org.blinkenlights.jid3.ID3Tag;
import org.blinkenlights.jid3.MP3File;
import org.blinkenlights.jid3.MediaFile;
import org.blinkenlights.jid3.io.TextEncoding;
import org.blinkenlights.jid3.v1.ID3V1Tag;
import org.blinkenlights.jid3.v1.ID3V1_0Tag;
import org.blinkenlights.jid3.v1.ID3V1_1Tag;
import org.blinkenlights.jid3.v2.*;

import java.io.File;
import java.io.FileInputStream;


/**
 * AudioTrack stores information about an audio.<br>
 * AudioTrack can update tags, rename file and delete itself (the file).<br>
 * @ToDo Use get/set for all fields
 */
public class AudioTrack implements Comparable, Cloneable {
    public enum TrackType {
        UNDEFINED,
        MP3,
        MP3_ID3V1,
        MP3_ID3V2,
        MP3_ID3V,
        OGG,
    }


    static public boolean   READ_ALL = true;
    public String           aFullFileName;
    public String           aFileName;
    public Long             aFileSize;
    public Long             aTimeSize;
    public String           aArtist;
    public String           aAlbum;
    public String           aTitle;
    public String           aNewTitle;
    public String           aComment;
    public String           aGenre;
    public String           aGenre2;
    public Integer          aBitRate;
    public Integer          aFreq;
    public String           aTime;
    public String           aYear;
    public String           aCopyrightInfo;
    public String           aComposer;
    public String           aEncodedBy;
    public String           aTrack;
    public String           aMode;
    public String           aVersion;
    public APICID3V2Frame[] aPictures;
    public boolean          aVbr;
    private boolean         aSelected;
    public boolean          aClearedOfPicture;
    public TrackType        aType;


    /**
     * Create object but DON'T read tags from file.
     *
     * @param name The filename
     */
    public AudioTrack(String name) {
        clear();
        aFileName = name;
    }


    /**
     * Create AudioTrack object and read tags from file.
     *
     * @param name          Filename
     * @param fullName      Filename and pathname
     */
    public AudioTrack(String name, String fullName) {
        clear();
        read(name, fullName);
    }


    /**
     * Clear all data.
     */
    public void clear() {
        setSelected(true);
        aClearedOfPicture = false;
        aFullFileName = "";
        aFileName = "";
        aFileSize = (long) 0;
        aTimeSize = (long) 0;
        aArtist = "";
        aAlbum = "";
        aTitle = "";
        aNewTitle = "";
        aComment = "";
        aGenre = "";
        aGenre2 = "";
        aBitRate = 0;
        aFreq = 0;
        aVersion = "";
        aMode = "";
        aVbr = false;
        aTime = "";
        aYear = "";
        aCopyrightInfo = "";
        aComposer = "";
        aEncodedBy = "";
        aTrack = "";
        aType = TrackType.UNDEFINED;
        aPictures = null;
    }


    /**
     * Clear all data.
     */
    public void clearID3v1() {
        aArtist = "";
        aAlbum = "";
        aTitle = "";
        aNewTitle = "";
        aComment = "";
        aGenre = "";
        aGenre2 = "";
        aYear = "";
        aCopyrightInfo = "";
        aComposer = "";
        aEncodedBy = "";
        aTrack = "";
    }


    /**
     * Clear all data for v2 tags.
     */
    public void clearID3v2() {
        aCopyrightInfo = "";
        aComposer = "";
        aEncodedBy = "";
    }


    /**
     * Create a copy of this track.
     *
     * @return  Copied object
     * @throws  CloneNotSupportedException
     */
    public Object clone() throws CloneNotSupportedException {
        super.clone();
        AudioTrack track = new AudioTrack("");
        track.aFullFileName = aFullFileName;
        track.aFileName = aFileName;
        track.aFileSize = aFileSize;
        track.aArtist = aArtist;
        track.aAlbum = aAlbum;
        track.aTitle = aTitle;
        track.aNewTitle = aNewTitle;
        track.aComment = aComment;
        track.aGenre = aGenre;
        track.aGenre2 = aGenre2;
        track.aBitRate = aBitRate;
        track.aFreq = aFreq;
        track.aVersion = aVersion;
        track.aVbr = aVbr;
        track.aMode = aMode;
        track.aTime = aTime;
        track.aTimeSize = aTimeSize;
        track.aYear = aYear;
        track.aCopyrightInfo = aCopyrightInfo;
        track.aComposer = aComposer;
        track.aEncodedBy = aEncodedBy;
        track.aTrack = aTrack;
        track.aType = aType;
        track.setSelected(isSelected());
        track.aPictures = aPictures;
        return track;
    }


    /**
     * Compare filename
     *
     * @param object    Another MP3 object
     * @return          0,1,-1
     */
    public int compareTo(Object object) {
        return aFileName.compareTo(((AudioTrack) object).aFileName);
    }


    /**
     * Delete file from disk.
     *
     * @throws Exception Throws exception if it fails
     */
    public void delete() throws Exception {
        File file = new File(aFullFileName);

        if (!file.exists()) {
            return;
        }

        if (!file.delete()) {
            throw new Exception(String.format("Can't delete file %s!", aFullFileName));
        }
    }

    public String getFullFileName() {
        return aFullFileName;
    }

    public String getFileName() {
        return aFileName;
    }

    public Long getFileSize() {
        return aFileSize;
    }

    public Long getTimeSize() {
        return aTimeSize;
    }

    public String getArtist() {
        return aArtist;
    }

    public String getAlbum() {
        return aAlbum;
    }

    public String getTitle() {
        return aTitle;
    }

    public String getNewTitle() {
        return aNewTitle;
    }

    public String getComment() {
        return aComment;
    }

    public String getGenre() {
        return aGenre;
    }

    public String getGenre2() {
        return aGenre2;
    }

    public Integer getBitRate() {
        return aBitRate;
    }

    public Integer getFreq() {
        return aFreq;
    }

    public String getTime() {
        return aTime;
    }

    public String getYear() {
        return aYear;
    }

    public String getCopyrightInfo() {
        return aCopyrightInfo;
    }

    public String getComposer() {
        return aComposer;
    }

    public String getEncodedBy() {
        return aEncodedBy;
    }

    public String getTrack() {
        return aTrack;
    }

    public String getMode() {
        return aMode;
    }

    public String getVersion() {
        return aVersion;
    }

    public APICID3V2Frame[] getPictures() {
        return aPictures;
    }

    public boolean isVbr() {
        return aVbr;
    }

    public boolean isSelected() {
        return aSelected;
    }

    public TrackType getType() {
        return aType;
    }

    /**
     * Get track tag write mode for an audio file
     *
     * @param fileName  Name of audio file
     * @return          Return the tag write mode AudioTrack.TrackType
     */
    public static TrackType getWriteTrackType(String fileName) {
        if (fileName.toLowerCase().endsWith(".mp3")) {
            String v1 = Pref.getPref(Constants.PREF_ID3V1_KEY, Constants.PREF_ID3V1_DEFAULT);
            String v2 = Pref.getPref(Constants.PREF_ID3V2_KEY, Constants.PREF_ID3V2_DEFAULT);

            if (v1.length() > 0 && v2.length() > 0) {
                return AudioTrack.TrackType.MP3_ID3V;
            }
            else if (v1.length() > 0 && v2.length() == 0) {
                return AudioTrack.TrackType.MP3_ID3V1;
            }
            else if (v1.length() == 0 && v2.length() > 0) {
                return AudioTrack.TrackType.MP3_ID3V2;
            }
            else if (v1.length() == 0 && v2.length() == 0) {
                return AudioTrack.TrackType.MP3;
            }
            else {
                assert false;
                return AudioTrack.TrackType.MP3;
            }
        }
        else {
            assert false;
            return AudioTrack.TrackType.UNDEFINED;
        }

    }


    /**
     * Read tags from file.
     *
     * @throws Exception Throws exception if it failes
     */
    public void read() {
        read(aFileName, aFullFileName);
    }


    /**
     * Read model from file
     *
     * @param name          Filename
     * @param fullName      Filename and pathname
     * @throws Exception    Throws exception if it failes
     */
    public void read(String name, String fullName) {
        clear();
        File file = new File(fullName);
        aFileName = name;
        aFullFileName = fullName;
        aFileSize = file.length();

        if (fullName.toLowerCase().endsWith(".mp3")) {
            boolean     istag1 = false, istag2 = false;
            MediaFile   track = new MP3File(file);

            try {
                ID3Tag[] tags = track.getTags();

                for (ID3Tag tag : tags) {
                    if (tag instanceof ID3V1_0Tag) {
                        istag1 = true;
                        readTag10(tag);
                    }
                    else if (tag instanceof ID3V1_1Tag) {
                        istag1 = true;
                        readTag11(tag);
                    }
                    else if (tag instanceof ID3V2_3_0Tag) {
                        istag2 = true;
                        readTag23(tag);
                    }
                }

                if (istag1 && istag2) {
                    aType = TrackType.MP3_ID3V;
                }
                else if (istag1) {
                    aType = TrackType.MP3_ID3V1;
                }
                else if (istag2) {
                    aType = TrackType.MP3_ID3V2;
                }
                else {
                    aType = TrackType.MP3;
                }

                if (READ_ALL) {
                    readAdditionalTag(aFullFileName);
                }
            }
            catch (Exception e) {
                if (Constants.DEBUG) e.printStackTrace();
            }
        }
    }


    /**
     * Read some extra meta data
     *
     * @param fileName      Filename
     * @throws Exception    Throws on fail
     */
    private void readAdditionalTag(String fileName) throws Exception {
        FileInputStream file    = new FileInputStream(fileName);
        Bitstream       in      = new Bitstream(file);
        Header          header  = in.readFrame();

        aBitRate    = header.bitrate() / 1000;
        aFreq       = header.frequency();
        aMode       = header.mode_string();
        aVersion    = header.version_string();
        aVbr        = header.vbr();
        aTimeSize   = (long) (header.total_ms(file.available()));
        aTime       = String.format("%02d:%02d", aTimeSize / 60000, aTimeSize / 1000 % 60);
        in.close();
    }


    /**
     * Parse MP3 1.0 tags
     *
     * @param tag   Tag object
     */
    private void readTag10(ID3Tag tag) {
        ID3V1_0Tag tag1 = (ID3V1_0Tag) tag;

        if (tag1.getArtist() != null) {
            aArtist = tag1.getArtist();
        }

        if (tag1.getAlbum() != null) {
            aAlbum = tag1.getAlbum();
        }

        if (tag1.getTitle() != null) {
            aTitle = tag1.getTitle();
        }

        if (tag1.getYear() != null) {
            aYear = tag1.getYear();
        }

        if (tag1.getGenre() != null) {
            aGenre = AudioGenre.get().parseGenre(tag1.getGenre().toString());
        }

        if (tag1.getComment() != null) {
            aComment = tag1.getComment();
        }
    }


    /**
     * Parse MP3 1.1 tags
     *
     * @param tag   Tag object
     */
    private void readTag11(ID3Tag tag) {
        ID3V1_1Tag tag1 = (ID3V1_1Tag) tag;

        if (tag1.getArtist() != null) {
            aArtist = tag1.getArtist();
        }

        if (tag1.getAlbum() != null) {
            aAlbum = tag1.getAlbum();
        }

        if (tag1.getTitle() != null) {
            aTitle = tag1.getTitle();
        }

        if (tag1.getYear() != null) {
            aYear = tag1.getYear();
        }

        if (tag1.getGenre() != null) {
            aGenre = AudioGenre.get().parseGenre(tag1.getGenre().toString());
        }

        if (tag1.getComment() != null) {
            aComment = tag1.getComment();
        }

        if (tag1.getAlbumTrack() > 0) {
            aTrack = Integer.toString(tag1.getAlbumTrack());
        }
    }


    /**
     * Parse MP3 2.3 tags
     *
     * @param tag           Tag object
     * @throws Exception    Throws on fail
     */
    private void readTag23(ID3Tag tag) {
        ID3V2_3_0Tag tag2 = (ID3V2_3_0Tag) tag;

        try {
            if (tag2.getArtist() != null) {
                aArtist = tag2.getArtist();
            }

            if (tag2.getAlbum() != null) {
                aAlbum = tag2.getAlbum();
            }

            if (tag2.getTitle() != null) {
                aTitle = tag2.getTitle();
            }

            if (tag2.getComment() != null) {
                aComment = tag2.getComment();
            }

            if (tag2.getGenre() != null) {
                String s = AudioGenre.get().parseGenre(tag2.getGenre());

                if (s.equalsIgnoreCase(AudioGenre.DEFAULT_STRING)) {
                    aGenre2 = tag2.getGenre();
                }
                else {
                    aGenre2 = s;
                }
            }

            try {
                aYear = Integer.toString(tag2.getYear());
            }
            catch (Exception e) {
                aYear = "";
            }

            try {
                aTrack = Integer.toString(tag2.getTrackNumber());
            }
            catch (Exception e) {
                aTrack = "";
            }

            if (tag2.getTCOPTextInformationFrame() != null) {
                aCopyrightInfo = String.format("%d %s", tag2.getTCOPTextInformationFrame().getCopyrightYear(), tag2.getTCOPTextInformationFrame().getCopyrightMessage());
            }

            if (tag2.getTENCTextInformationFrame() != null) {
                aEncodedBy = tag2.getTENCTextInformationFrame().getEncodedBy();
            }

            if (tag2.getTCOMTextInformationFrame() != null) {
                String[] composers = tag2.getTCOMTextInformationFrame().getComposers();

                for (String s : composers) {
                    if (aComposer.length() > 0) {
                        aComposer += " / ";
                    }
                    aComposer += s;
                }
            }

            aPictures = tag2.getAPICFrames();
        }
        catch (Exception e) {
            Log.get().addTime(1, e.getMessage());
        }
    }


    /**
     * BaseRename file.<br>
     * The new file will be in the same directory.
     *
     * @param newName       The new filename
     * @throws Exception    Throws on fail
     */
    public void rename(String newName) throws Exception {
        if (aFileName.compareTo(newName) == 0) {
            return;
        }

        File from = new File(aFullFileName);
        File to = new File(from.getParent() + File.separator + newName);

        if (!from.exists()) {

            throw new Exception(String.format("File %s is missing!", aFileName));
        }

        if (to.exists()) {
            throw new Exception(String.format("File %s exist already!", newName));
        }

        if (!from.renameTo(to)) {
            throw new Exception(String.format("Renaming %s to %s failed!", aFileName, newName));
        }

        aFileName = newName;
        aFullFileName = from.getParent() + File.separator + newName;
    }


    public void setFullFileName(String aFullFileName) {
        this.aFullFileName = aFullFileName;
    }

    public void setFileName(String aFileName) {
        this.aFileName = aFileName;
    }

    public void setFileSize(Long aFileSize) {
        this.aFileSize = aFileSize;
    }

    public void setTimeSize(Long aTimeSize) {
        this.aTimeSize = aTimeSize;
    }

    public void setArtist(String aArtist) {
        this.aArtist = aArtist;
    }

    public void setAlbum(String aAlbum) {
        this.aAlbum = aAlbum;
    }

    public void setTitle(String aTitle) {
        this.aTitle = aTitle;
    }

    public void setNewTitle(String aNewTitle) {
        this.aNewTitle = aNewTitle;
    }

    public void setComment(String aComment) {
        this.aComment = aComment;
    }

    public void setGenre(String aGenre) {
        this.aGenre = aGenre;
    }

    public void setGenre2(String aGenre2) {
        this.aGenre2 = aGenre2;
    }

    public void setBitRate(Integer aBitRate) {
        this.aBitRate = aBitRate;
    }

    public void setFreq(Integer aFreq) {
        this.aFreq = aFreq;
    }

    public void setTime(String aTime) {
        this.aTime = aTime;
    }

    public void setYear(String aYear) {
        this.aYear = aYear;
    }

    public void setCopyrightInfo(String aCopyrightInfo) {
        this.aCopyrightInfo = aCopyrightInfo;
    }

    public void setComposer(String aComposer) {
        this.aComposer = aComposer;
    }

    public void setEncodedBy(String aEncodedBy) {
        this.aEncodedBy = aEncodedBy;
    }

    public void setTrack(String aTrack) {
        this.aTrack = aTrack;
    }

    public void setMode(String aMode) {
        this.aMode = aMode;
    }

    public void setVersion(String aVersion) {
        this.aVersion = aVersion;
    }

    public void setPictures(APICID3V2Frame[] aPictures) {
        this.aPictures = aPictures;
    }

    public void setVbr(boolean aVbr) {
        this.aVbr = aVbr;
    }

    public void setSelected(boolean aSelected) {
        this.aSelected = aSelected;
    }

    public void setType(TrackType aType) {
        this.aType = aType;
    }

    /**
     * Set image.
     *
     * @param pictureModel  Picturemodel contains the binary image data
     * @param label         Picture description
     * @throws ID3Exception Exception if it fails
     */
    public void setImage(PictureModel pictureModel, String label) throws ID3Exception {
        if (pictureModel.getPictureData() == null) { // Clear image data if input data is null
             aPictures = null;
        }
        else if (aPictures != null && aPictures.length > 0){ // At least one image exist so set new first image
            aPictures[0].setDescription(label);
            aPictures[0].setPictureData(pictureModel.getPictureData());
        }
        else if (pictureModel.getPictureData() != null) { // No pictures were defined, add a new one
            APICID3V2Frame frame = new APICID3V2Frame(pictureModel.getMimeType().toString(), APICID3V2Frame.PictureType.FrontCover, "", pictureModel.getPictureData());
            frame.setDescription(label);
            aPictures = new APICID3V2Frame[]{frame};
        }
    }


    /**
     * Change encoding to unicode or iso8859.<br>
     * Which depends on application settings.<br>
     * This is from a suggestion from Hindin Joseph<br>
     * It doesn't check is strings contains unicode characters.
     */
    public void setupEncoding() {
        if (Pref.getPref(Constants.PREF_UNICODE_KEY, "").equals("YES")) {
            TextEncoding.setDefaultTextEncoding(TextEncoding.UNICODE);
        }
        else {
            TextEncoding.setDefaultTextEncoding(TextEncoding.ISO_8859_1);
        }
    }


    /**
     * Update tags in file.
     *
     * @param type          Tag type
     * @throws Exception    Throws on fail
     */
    public void write(TrackType type) throws Exception {
        aGenre = AudioGenre.get().parseGenre(aGenre);

        if (type == TrackType.MP3 || type == TrackType.MP3_ID3V || type == TrackType.MP3_ID3V1 || type == TrackType.MP3_ID3V2) {
            setupEncoding();

            boolean         istag1 = false;
            boolean         istag2 = false;
            MediaFile       track = new MP3File(new File(aFullFileName));
            ID3V1_1Tag      tag1 = new ID3V1_1Tag();
            ID3V2_3_0Tag    tag2 = new ID3V2_3_0Tag();

            try {
                if (type == AudioTrack.TrackType.MP3_ID3V || type == AudioTrack.TrackType.MP3_ID3V1) {
                    istag1 = true;
                    tag1.setAlbum(aAlbum);
                    tag1.setArtist(aArtist);
                    tag1.setTitle(aTitle);
                    tag1.setYear(aYear);
                    tag1.setComment(aComment);
                    tag1.setGenre(ID3V1Tag.Genre.lookupGenre(aGenre));
                    if (Util.getNumber(aTrack, -1) >= 0) tag1.setAlbumTrack(Util.getNumber(aTrack, -1));
                }

                if (type == AudioTrack.TrackType.MP3_ID3V || type == AudioTrack.TrackType.MP3_ID3V2) {
                    istag2 = true;
                    tag2.setAlbum(aAlbum);
                    tag2.setArtist(aArtist);
                    tag2.setTitle(aTitle);
                    if (Util.getNumber(aYear, -1) >= 0) tag2.setYear(Util.getNumber(aYear, -1));
                    tag2.setComment(aComment);
                    tag2.setGenre(aGenre2);
                    if (Util.getNumber(aTrack, -1) >= 0) tag2.setTrackNumber(Util.getNumber(aTrack, -1));

                    if (aCopyrightInfo.length() > 5) {
                        String  y = aCopyrightInfo.substring(0, 4);
                        String  m = aCopyrightInfo.substring(5, aCopyrightInfo.length());
                        int     y2 = Util.getNumber(y, 2000);

                        tag2.setTCOPTextInformationFrame(new TCOPTextInformationID3V2Frame(y2, m));
                    }

                    tag2.setTCOMTextInformationFrame(new TCOMTextInformationID3V2Frame(aComposer));
                    tag2.setTENCTextInformationFrame(new TENCTextInformationID3V2Frame(aEncodedBy));

                    if (Pref.getPref(Constants.PREF_USEPIC_KEY, Constants.PREF_USEPIC_DEFAULT).length() > 0 && aPictures != null) {
                        for (APICID3V2Frame pic : aPictures) {
                            tag2.addAPICFrame(pic);
                        }
                    }
                }

                if (istag1 && istag2) {
                    aType = TrackType.MP3_ID3V;
                    track.setID3Tag(tag1);
                    track.setID3Tag(tag2);
                }
                else if (istag1) {
                    clearID3v2();
                    aType = TrackType.MP3_ID3V1;
                    track.removeID3V2Tag();
                    track.setID3Tag(tag1);
                }
                else if (istag2) {
                    aType = TrackType.MP3_ID3V2;
                    track.removeID3V1Tag();
                    track.setID3Tag(tag2);
                }
                else {
                    clearID3v1();
                    track.removeTags();
                    aType = TrackType.MP3;
                }

                aClearedOfPicture = false;
                track.sync();
            }
            catch (Exception e) {
                try {
                    read();
                }
                catch (Exception e1) {
                    if (Constants.DEBUG) e.printStackTrace();
                }

                throw new Exception(e.getMessage());
            }
        }
    }
}
