/*
* Copyright 2004 - 2009 by dronten@gmail.com
*
* This source is distributed under the terms of the GNU PUBLIC LICENSE version 3
* http://www.gnu.org/licenses/gpl.html
*/


package com.googlepages.dronten.jtagger.gui.leftpanel;

import javax.swing.*;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.Arrays;


/**
 * A director browser panel.
 * A directory path can be restored at startup.
 */
public class DirTree extends JPanel implements TreeExpansionListener, ActionListener, TreeSelectionListener {
    private static final long serialVersionUID = 1;

    /**
     * Store directory data for the DirTree.
     */
    class DirNode {
        private File aFile;


        /**
         * Create new DirNode object.
         *
         * @param file - A directory to show in the tree
         */
        public DirNode(File file) {
            aFile = file;
        }


        /**
         * Return file object.
         *
         * @return
         */
        public File getFile() {
            return aFile;
        }


        /**
         * Set file object.
         *
         * @param file
         */
        public void setFile(File file) {
            aFile = file;
        }


        /**
         * Return the shortest path name.
         *
         * @return
         */
        public String toString() {
            if (aFile.getName().length() == 0) {
                return aFile.getPath();
            }
            else {
                return aFile.getName();
            }
        }
    }


    private static String           MENU_REFRESH = "Reload selected directory";
    private static String           MENU_REFRESH2 = "Reload parent directory";
    private static String           MENU_RENAME = "Rename directory";
    private static String           MENU_CREATE = "Create directory";
    private static String           MENU_DELETE = "Delete directory";
    private static String           ASK_RENAME = "Enter new directory name";
    private static String           TITLE_RENAME = "Rename FileNames";
    private static String           ASK_DELETE = "Delete this director (must be empty)?";
    private static String           TITLE_DELETE = "Delete FileNames";
    private static String           ASK_CREATE = "Enter new directory name";
    private static String           TITLE_CREATE = "Create FileNames";


    private DefaultMutableTreeNode  aTop = null;
    private DefaultMutableTreeNode  aSelected = null;
    private JPopupMenu              aMenu = null;
    private DefaultTreeModel        aTreeModel = null;
    private JTree                   aTree = null;
    private DirListener             aDirListener = null;


    /**
     * @param rootName
     */
    public DirTree(String rootName) {
        super();

        aTop = new DefaultMutableTreeNode(rootName);
        aTreeModel = new DefaultTreeModel(aTop);
        aTree = new JTree(aTreeModel);
        JScrollPane treeView = new JScrollPane(aTree);

        setLayout(new BorderLayout());
        add(treeView, BorderLayout.CENTER);

        aTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        aTree.setShowsRootHandles(true);
        aTree.addTreeExpansionListener(this);
        aTree.addTreeSelectionListener(this);
        aTree.setEditable(false);
        Icon icon = ((DefaultTreeCellRenderer) aTree.getCellRenderer()).getClosedIcon();
        ((DefaultTreeCellRenderer) aTree.getCellRenderer()).setLeafIcon(icon);
        aMenu = new JPopupMenu("");

        JMenuItem menuItem = new JMenuItem(MENU_REFRESH);
        menuItem.addActionListener(this);
        aMenu.add(menuItem);

        menuItem = new JMenuItem(MENU_CREATE);
        menuItem.addActionListener(this);
        aMenu.add(menuItem);

        menuItem = new JMenuItem(MENU_RENAME);
        menuItem.addActionListener(this);
        aMenu.add(menuItem);

        menuItem = new JMenuItem(MENU_DELETE);
        menuItem.addActionListener(this);
        aMenu.add(menuItem);

        MouseListener popupListener = new PopupListener(this);
        aTree.addMouseListener(popupListener);
    }


    /**
     * Callback for popup menu.
     *
     * @param e
     */
    @Override public void actionPerformed(ActionEvent e) {
        if (aSelected == null || aSelected == aTop) {
            return;
        }

        if (e.getActionCommand().equalsIgnoreCase(MENU_REFRESH)) {
            if (aSelected != null) {
                addDir(aSelected);
                addSubDirs(aSelected);
                aTreeModel.reload(aSelected);
                aDirListener.valueChanged(((DirNode) aSelected.getUserObject()).getFile().getPath());
            }
        }
        else if (e.getActionCommand().equalsIgnoreCase(MENU_RENAME)) {
            if (aSelected != null) {
                String      newDirName = JOptionPane.showInputDialog(this, ASK_RENAME, TITLE_RENAME, JOptionPane.QUESTION_MESSAGE);
                DirNode     dirNode = (DirNode) aSelected.getUserObject();

                if (newDirName != null && newDirName.length() > 0 && newDirName.compareTo(dirNode.getFile().getName()) != 0) {
                    File from = dirNode.getFile();
                    File to = new File(from.getParent() + File.separator + newDirName);
                    if (!from.renameTo(to)) {
                        Toolkit.getDefaultToolkit().beep();
                    }
                    else {
                        dirNode.setFile(to);
                        aTreeModel.reload(aSelected);
                        aDirListener.valueChanged(((DirNode) aSelected.getUserObject()).getFile().getPath());
                    }
                }
            }
        }
        else if (e.getActionCommand().equalsIgnoreCase(MENU_CREATE)) {
            if (aSelected != null) {
                String      newDirName = JOptionPane.showInputDialog(this, ASK_CREATE, TITLE_CREATE, JOptionPane.QUESTION_MESSAGE);
                DirNode     dirNode = (DirNode) aSelected.getUserObject();

                if (newDirName.length() > 0 && newDirName.compareTo(dirNode.getFile().getName()) != 0) {
                    File currDir = dirNode.getFile();
                    File newDir = new File(currDir.getPath() + File.separator + newDirName);
                    if (!newDir.mkdir()) {
                        Toolkit.getDefaultToolkit().beep();
                    }
                    else {
                        aSelected.removeAllChildren();
                        addDir(aSelected);
                        aTreeModel.reload(aSelected);
                        aDirListener.valueChanged(((DirNode) aSelected.getUserObject()).getFile().getPath());
                    }
                }
            }
        }
        else if (e.getActionCommand().equalsIgnoreCase(MENU_DELETE)) {
            if (aSelected != null) {
                int         answer = JOptionPane.showConfirmDialog(this, ASK_DELETE, TITLE_DELETE, JOptionPane.YES_NO_OPTION);
                DirNode     dirNode = (DirNode) aSelected.getUserObject();

                if (answer == JOptionPane.YES_OPTION) {
                    File del = dirNode.getFile();

                    if (!del.delete()) {
                        Toolkit.getDefaultToolkit().beep();
                    }
                    else {
                        aSelected = (DefaultMutableTreeNode) aSelected.getParent();
                        addDir(aSelected);
                        addSubDirs(aSelected);
                        aTreeModel.reload(aSelected);
                        aDirListener.valueChanged(((DirNode) aSelected.getUserObject()).getFile().getPath());
                    }
                }
            }
        }
    }


    /**
     * Add child directories to parent directory.
     *
     * @param node
     */
    private void addDir(DefaultMutableTreeNode node) {
        if (node == null || node == aTop) {
            return;
        }

        node.removeAllChildren();
        Object nodeInfo = node.getUserObject();
        addDir2(node, ((DirNode) nodeInfo).getFile());
    }


    /**
     * Add child directories to parent directory.
     *
     * @param node
     */
    private void addDir2(DefaultMutableTreeNode node, File f) {
        File files[] = null;

        try {
            if (f.isDirectory()) {
                files = f.listFiles();
            }
        }
        catch (Exception e) {
        }

        if (files != null) {
            Arrays.sort(files);

            for (File file : files) {
                if (file.isDirectory()) {
                    node.add(new DefaultMutableTreeNode(new DirNode(file)));
                }
            }
        }
    }


    /**
     * Add child directories for child directories.
     *
     * @param node
     */
    private void addSubDirs(DefaultMutableTreeNode node) {
        for (int f = 0; f < node.getChildCount(); f++) {
            DefaultMutableTreeNode currentNode = (DefaultMutableTreeNode) node.getChildAt(f);
            if (currentNode.isLeaf()) {
                addDir(currentNode);
            }
        }
    }


    /**
     * Set directory listener.
     *
     * @param dirListener
     */
    public void addDirListener(DirListener dirListener) {
        aDirListener = dirListener;
    }


    /**
     * @return
     */
    public JTree getTree() {
        return aTree;
    }


    /**
     * Read all root entries in the file system.
     */
    public void initRoot() {
        File files[];

        files = File.listRoots();
        aTop.removeAllChildren();
        for (File file : files) {
            DefaultMutableTreeNode node = new DefaultMutableTreeNode(new DirNode(file));
            aTop.add(node);
        }
    }


    /**
     * @param path
     */
    public void restore(String path) {
        String[] paths = path.split("\\" + File.separator);
        DefaultMutableTreeNode node = null;

        if (paths.length > 0) {
            for (int f = 0; f < aTop.getChildCount(); f++) {
                DefaultMutableTreeNode  currentNode = (DefaultMutableTreeNode) aTop.getChildAt(f);
                DirNode                 dirNode = (DirNode) currentNode.getUserObject();

                if ((paths[0] + File.separator).equalsIgnoreCase(dirNode.getFile().getPath())) {
                    node = DirTree.restore(paths, 1, currentNode, this);
                    break;
                }
            }
        }

        if (node != null) {
            aTree.setSelectionPath(new TreePath(node.getPath()));
            aTree.scrollPathToVisible(new TreePath(node.getPath()));
        }
    }


    /**
     * @param paths
     * @param pos
     * @param node
     * @param tree
     * @return
     */
    private static DefaultMutableTreeNode restore(String[] paths, int pos, DefaultMutableTreeNode node, DirTree tree) {
        DefaultMutableTreeNode retNode = node;

        if (paths.length > pos) {
            if (node.isLeaf()) {
                tree.addDir(node);
            }

            for (int f = 0; f < node.getChildCount(); f++) {
                DefaultMutableTreeNode currentNode = (DefaultMutableTreeNode) node.getChildAt(f);
                DirNode dirNode = (DirNode) currentNode.getUserObject();

                if ((paths[pos]).equalsIgnoreCase(dirNode.getFile().getName())) {
                    retNode = DirTree.restore(paths, pos + 1, currentNode, tree);
                    break;
                }
            }
        }

        return retNode;
    }

    /**
     * @param event
     */
    @Override public void treeCollapsed(TreeExpansionEvent event) {
    }


    /**
     * Add child dir to all child dir of current opened directory.
     *
     * @param event
     */
    @Override public void treeExpanded(TreeExpansionEvent event) {
        //DefaultMutableTreeNode dirNode = (DefaultMutableTreeNode) event.getPath().getLastPathComponent();

        //addSubDirs(dirNode);
    }


    /**
     * @param treeSelectionEvent
     */
    public void valueChanged(TreeSelectionEvent treeSelectionEvent) {
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) aTree.getLastSelectedPathComponent();
        aSelected = node;

        if (node != null && node != aTop && aDirListener != null) {
            addDir(aSelected);
            aDirListener.valueChanged(((DirNode) aSelected.getUserObject()).getFile().getPath());
        }
    }

    /**
     * Show popup menu
     */
    private class PopupListener extends MouseAdapter {
        DirTree aTree;


        /**
         *
         * @param tree
         */
        PopupListener(DirTree tree) {
            aTree = tree;
        }


        /**
         *
         * @param e
         */
        public void mousePressed(MouseEvent e) {
            maybeShowPopup(e);
        }


        /**
         *
         * @param e
         */
        public void mouseReleased(MouseEvent e) {
            maybeShowPopup(e);
        }


        /**
         * 
         * @param e
         */
        private void maybeShowPopup(MouseEvent e) {
            if (e.isPopupTrigger()) {
                aTree.aMenu.show(e.getComponent(), e.getX(), e.getY());
            }
        }
    }
}
